#!/usr/local/bin/ruby
#
# csv-parser#1 - normal version.
# 
# @author yosuke hara
# @version 2009-05-14
#
module CSV
  module Parser
    module_function
    def parse(source="")
      return nil if source.empty?

      records = []
      source.chomp!
      source.scan(/(([^",][^,]*)|"(([^"]|"")*)")/) do |m0, m1, m2|
        records << (
          m2 ? m2.gsub(/""/, '"') :
          m1 ? m1 : ""
        )
      end
      return records
    end
  end
  
  module Base
    class CouldNotFindRecord < StandardError
    end
    
    class CSVRecord
      COLUMN_ROW_NUM = 1
      
      def initialize(path = "")
        @columns = []
        @records = []
        @path = path
        self.load
      end
      
      def find(*args)
        return nil if @records.empty?

        case args.first
          when :first then find_initial
          when :last  then find_last
          when :all   then find_all
          else             find_by_ids(args)
        end
      end

    protected
      def load
        File.open(@path, "r") {|file|
          row_num = 0
          file.each {|line|
            set_values((row_num += 1), CSV::Parser::parse(line))
          }
        }
      end
      
      def set_values(row_num, values)
      end

    private
      COL_ID = "id"
      
      def find_initial
        return nil if @records.empty?
        @records.first
      end

      def find_last
        return nil if @records.empty?
        @records.last
      end

      def find_all
        @records
      end

      def find_by_ids(*ids)
        expects_array = ids.first.kind_of?(Array)
        return ids.first if expects_array && ids.first.empty?

        begin
          ids = ids.flatten.compact.uniq
          case ids.size
            when 0 then raise CouldNotFindRecord
            when 1 then find_one(ids.first)
            else        find_some(ids)
          end
        rescue => e
          p e.to_s
        end
      end

      def find_one(id)
        result = nil
        @records.each do |value|
          value_id = (value.instance_of? Hash) ? value[COL_ID].to_i : value.id
          
          if value_id == id
            result = value
            break
          end
        end
        raise CouldNotFindRecord unless result
        return result
      end

      def find_some(ids)
        result = []
        ids.each do |id|
          result << find_one(id)
        end
        raise CouldNotFindRecord if result.empty?
        return result
      end
    end
  end
end

#
# for Contact CSV File.
#
class Contact < CSV::Base::CSVRecord
  COL_ID         = 0
  COL_FIRST_NAME = 1
  COL_LAST_NAME  = 2
  COL_EMAIL      = 3
  
  def set_values(row_num, values)
    return if row_num == COLUMN_ROW_NUM
    
    # contact value object - nameless class.
    contact_class = Class.new{
      attr_accessor :id, :first_name, :last_name, :email
      
      def to_s
        str = ""
        str << "id:#{self.id}, "
        str << "first_name:#{self.first_name}, "
        str << "last_name:#{self.last_name}, "
        str << "email:#{self.email}"
      end
    }
    
    contact = contact_class.new
    contact.id         = values[COL_ID].to_i
    contact.first_name = values[COL_FIRST_NAME]
    contact.last_name  = values[COL_LAST_NAME]
    contact.email      = values[COL_EMAIL]
    @records << contact
  end
end

#
# for CSV File.
#
class CSVHash < CSV::Base::CSVRecord
  def set_values(row_num, values)
    if (row_num == 1)
      @columns = values
    else
      hash = {}

      values.length.times do |i|
        hash[@columns[i]] = values[i]
      end
      @records << hash
    end
  end
end

#
# Test-Codes. (require: RSpec-1.2.5+)
#
describe Contact do
  before do
    @contact = Contact.new "Contact.csv"
  end

  it "should find first record" do
    contact = @contact.find(:first)
    contact.id.should         == 1
    contact.first_name.should == "Miles"
    contact.last_name.should  == "Davis"
    contact.email.should      == "miles.davis@jazz.com"
    p contact.to_s
  end

  it "should find last record" do
    contact = @contact.find(:last)
    contact.id.should         == 7
    contact.first_name.should == "Chic"
    contact.last_name.should  == "Corea"
    contact.email.should      == "chic.corea@jazz.com"
    p contact.to_s
  end
  
  it "should find all records" do
    contacts = @contact.find(:all)
    contacts.length.should == 7
    p contacts
  end
  
  it "should not find record" do
    contact = @contact.find(0)
    contact.should == nil
  end
  
  it "should find by id" do
    contact = @contact.find(3)
    contact.id.should         == 3
    contact.first_name.should == "John"
    contact.last_name.should  == "Coltrane"
    contact.email.should      == "john.coltrane@jazz.com"
    p contact.to_s
  end
  
  it "should find by ids" do
    contacts = @contact.find([1,3,5,7])
    contacts.length.should == 4
    p contacts
  end
end


describe CSVHash do
  before do
    @hash = CSVHash.new "Contact.csv"
  end

  it "should find first record" do
    contact = @hash.find(:first)
    contact["id"].to_i.should    == 1
    contact["first_name"].should == "Miles"
    contact["last_name"].should  == "Davis"
    contact["email"].should      == "miles.davis@jazz.com"
    p contact
  end

  it "should find last record" do
    contact = @hash.find(:last)
    contact["id"].to_i.should    == 7
    contact["first_name"].should == "Chic"
    contact["last_name"].should  == "Corea"
    contact["email"].should      == "chic.corea@jazz.com"
    p contact
  end
  
  it "should find all records" do
    contacts = @hash.find(:all)
    contacts.length.should == 7
    p contacts
  end
  
  it "should not find record" do
    contact = @hash.find(0)
    contact.should == nil
  end
  
  it "should find by id" do
    contact = @hash.find(3)
    contact["id"].to_i.should    == 3
    contact["first_name"].should == "John"
    contact["last_name"].should  == "Coltrane"
    contact["email"].should      == "john.coltrane@jazz.com"
    p contact
  end
  
  it "should find by ids" do
    contacts = @hash.find([1,3,5,7])
    contacts.length.should == 4
    p contacts
  end
end
