#!/usr/local/bin/ruby
#
# csv-parser#2 - Inner-DSL version.
#
# @author yosuke hara
# @version 2009-05-14
#
module CSV
  module Parser
    module_function
    def parse(source="")
      return nil if source.empty?

      records = []
      source.chomp!
      source.scan(/(([^",][^,]*)|"(([^"]|"")*)")/) do |m0, m1, m2|
        records << (
          m2 ? m2.gsub(/""/, '"') :
          m1 ? m1 : ""
        )
      end
      return records
    end
  end
  
  module Base
    class FileNotFind < StandardError
    end
    class CouldNotFindRecord < StandardError
    end

    class CSVRecord
      CSV_SUFFIX = ".csv"
      COLUMN_ROW_NUM = 1
      COL_ID = "id"

      def initialize()
        @columns = []
        @records = []
        @path = self.class.to_s.concat(CSV_SUFFIX)
        @klass = Class.new
        
        self.load
      end

      def find(*args)
        return nil if @records.empty?

        case args.first
          when :first then find_initial
          when :last  then find_last
          when :all   then find_all
          else             find_by_ids(args)
        end
      end

    protected
      def load
        raise FileNotFind unless File.exist? @path

        File.open(@path, "r") {|file|
          row_num = 0

          file.each {|line|
            columns = CSV::Parser::parse(line)

            if (row_num += 1) == COLUMN_ROW_NUM
              generate_accessors columns
              
              to_string = ""
              len = columns.length
              len.times do |i|
                col = columns[i]
                to_string << "\"#{col}:\#{self.#{col}}"
                to_string << (len - 1 == i ? "\"" : ", \"")
              end

              @klass.class_eval %{
                def to_s
                  #{to_string}
                end
              }
            else
              set_values columns
            end
          }
        }
      end

      def generate_accessors(columns)
        @columns = columns
        @columns.each do |col|
          @klass.class_eval %{
            def #{col}
              @#{col}
            end

            def #{col}= (value)
              @#{col} = value
            end
          }
        end
      end

      def set_values(values)
        object = @klass.new

        values.length.times do |i|
          if @columns[i] == COL_ID
            object.instance_variable_set("@#{@columns[i]}", values[i].to_i)
          else
            object.instance_variable_set("@#{@columns[i]}", values[i])
          end
        end

        @records << object
      end

    private
      def find_initial
        return nil if @records.empty?
        @records.first
      end

      def find_last
        return nil if @records.empty?
        @records.last
      end

      def find_all
        @records
      end

      def find_by_ids(*ids)
        expects_array = ids.first.kind_of?(Array)
        return ids.first if expects_array && ids.first.empty?

        begin
          ids = ids.flatten.compact.uniq
          case ids.size
            when 0
              raise CouldNotFindRecord
            when 1
              find_one(ids.first)
            else
              find_some(ids)
            end
        rescue => e
          p e.to_s
        end
      end

      def find_one(id)
        result = nil
        @records.each do |value|
          value_id = (value.instance_of? Hash) ? value[COL_ID].to_i : value.id
          
          if value_id == id
            result = value
            break
          end
        end
        raise CouldNotFindRecord unless result
        return result
      end

      def find_some(ids)
        result = []
        ids.each do |id|
          result << find_one(id)
        end
        raise CouldNotFindRecord if result.empty?
        return result
      end
    end
  end
end

#
#
#
class Contact < CSV::Base::CSVRecord
end

#
# Test-Codes. (require: RSpec-1.2.5+)
#
describe Contact do
  before do
    @contact = Contact.new
  end

  it "should find first record" do
    contact = @contact.find(:first)
    contact.id.should         == 1
    contact.first_name.should == "Miles"
    contact.last_name.should  == "Davis"
    contact.email.should      == "miles.davis@jazz.com"
    p contact.to_s
  end

  it "should find last record" do
    contact = @contact.find(:last)
    contact.id.should         == 7
    contact.first_name.should == "Chic"
    contact.last_name.should  == "Corea"
    contact.email.should      == "chic.corea@jazz.com"
    p contact.to_s
  end
  
  it "should find all records" do
    contacts = @contact.find(:all)
    contacts.length.should == 7
    p contacts
  end
  
  it "should not find record" do
    contact = @contact.find(0)
    contact.should == nil
  end
  
  it "should find by id" do
    contact = @contact.find(3)
    contact.id.should         == 3
    contact.first_name.should == "John"
    contact.last_name.should  == "Coltrane"
    contact.email.should      == "john.coltrane@jazz.com"
    p contact.to_s
  end
  
  it "should find by ids" do
    contacts = @contact.find([1,3,5,7])
    contacts.length.should == 4
    p contacts
  end
end